import os
from telethon import events


class Handler:
    """Base class for the telegrambot event handler."""

    def __init__(self, bot):
        self.directory = os.getcwd()
        self.directory_name = os.path.relpath(self.directory, os.pardir)
        self._bot = bot
        self.setup()
        self._add_subscribtions()

    def setup(self):
        """Setup children."""
        pass

    def _add_subscribtions(self):
        """Subscribe to the different events."""
        # should be overloaded in childclass.
        return

    def _subscribe(self, pattern, callback):
        """Subscribe to new messages that follow a certain pattern"""
        self._bot.add_event_handler(callback, events.NewMessage(pattern=pattern))



import glob
import os
from f90nml import Parser

from .handler import Handler


class MITgcmDataParser(Parser):
    """Parser for MITgcm data files."""

    def __init__(self):
        super().__init__()
        self.comment_tokens += '#'
        self.end_comma = True
        self.indent = " "
        self.column_width = 72
        self.sparse_arrays = True


def get_parameter(datafile, keyword):
    """
    Function to parse the MITgcm 'data' file and return the parameter values
    of the given specific keyword.

    Parameters
    ----------
    datafile: string
        Full path to the MITgcm data file.
    keyword: string
        Parameter of which the value is required.

    Returns
    ----------
    value: string
        The value associated with the given keyword is returned as a string (!).
    """

    if not os.path.isfile(datafile):
        raise FileNotFoundError("could not find the datafile.")

    parser = MITgcmDataParser()
    data = parser.read(datafile)

    for section in data:
        for key, val in data[section].items():
            if key.lower() == keyword.lower():
                return val

    raise KeyError("Keyword not found")


class MITgcmHandler(Handler):
    """Handler for a MITgcm simulation."""

    def setup(self):
        if "run" not in os.listdir(self.directory):
            raise FileNotFoundError("Your current working directory needs to contain a run directory")

        self.run_dir = os.path.join(self.directory, "run")

        if "input" not in os.listdir(self.directory):
            raise FileNotFoundError("Your current working directory needs to contain an input directory")

        if "data" not in os.listdir(os.path.join(self.directory, "input")):
            raise FileNotFoundError("You need a data file inside your input directory")

        self.datafile = os.path.join(os.path.join(os.getcwd(), "input"), "data")

    def _add_subscribtions(self):
        """Subscribe to the different events."""
        self._subscribe(f'(?i)status (?i){self.directory_name}', self.respond_status)
        self._subscribe(f'(?i)progress (?i){self.directory_name}', self.respond_progress)
        self._subscribe(f'(?i)error (?i){self.directory_name}', self.respond_error)
        self._subscribe(f'(?i)list', self.respond_all)

    async def respond_all(self, event):
        """Answer the directory name upon asking for a list of active simulations."""
        return await event.respond(f"{self.directory_name}")

    async def respond_progress(self, event):
        """Answer the progress of this simulation."""
        return await event.respond(f"We are at {self.progress}%.")

    async def respond_error(self, event):
        """Answer the errors that are in the STDERR.* files."""
        errors = ""
        for stderr in glob.glob(os.path.join(self.run_dir, "STDERR.*")):
            with open(stderr, "r") as stderr_f:
                for line in stderr_f:
                    errors += f"{line}\n" if "INI_PARMS" not in line else ""

        if not bool(errors):
            return await event.respond("No errors yet.")

        return await event.respond(f"We got the following errors:\n{errors}")

    async def respond_status(self, event):
        """Answer a status update."""
        last_iter = self.last_iter
        if not last_iter:
            return await event.respond("No outputs yet.")

        return await event.respond(
            "Highest iteration is {:d} which translates to {:.1f} days.".format(last_iter, self.last_day))

    @property
    def last_day(self):
        """Return the time (in days) of current snapshot."""
        deltat = int(get_parameter(self.datafile, "deltat"))
        return deltat * self.last_iter / 3600 / 24

    @property
    def last_iter(self):
        """Return the iterationnumber of current snapshot."""
        iters = [int(file.split(".")[1]) for file in glob.glob(os.path.join(self.run_dir, "T.*.meta"))]
        if bool(iters):
            return max(iters)

    @property
    def progress(self):
        """Return the progress of the simulation (in percent)."""
        nIter0 = int(get_parameter(self.datafile, "nIter0"))
        nTimeSteps = int(get_parameter(self.datafile, "nTimeSteps"))
        if bool(self.last_iter):
            return (self.last_iter - nIter0) / nTimeSteps * 100
        else:
            return 0.0



import os
import yaml
from appdirs import AppDirs
from argparse import ArgumentParser

CONF_API_ID = "api_id"
CONF_API_HASH = "api_hash"
CONF_BOT_TOKEN = "bot_token"


def get_config_path():
    """Set the config path."""
    return AppDirs("telegram_simulation_bot").user_config_dir + ".yaml"


def safe_config(config):
    """Safe the config file."""
    with open(get_config_path(), "w") as f:
        yaml.dump(config, f, Dumper=yaml.Dumper)


def read_config():
    """Read the config file."""
    if not os.path.isfile(get_config_path()):
        raise FileNotFoundError("We did not find a config for telegram_simulation_bot.")

    with open(os.path.join(get_config_path())) as f:
        return yaml.load(f, Loader=yaml.FullLoader)


def delete_config():
    """Delete the config file."""
    print("deleting the configuration.")
    os.remove(get_config_path())


def setup(args):
    """Carry out the configuration."""

    if args.delete:
        delete_config()
        quit()

    if args.config or not os.path.isfile(get_config_path()):
        config = {}
        print("-- telegram_simulation_bot --")
        print("starting the configuration.\n")
        print("Start a chat with the BotFather (https://telegram.me/BotFather) and create your bot.\n")
        config[CONF_BOT_TOKEN] = str(input("Input a bot token:\n>"))
        print("\nGet an api_id and api_hash from https://my.telegram.org, under API Development.\n"
              "More info: https://docs.telethon.dev/en/latest/basic/signing-in.html\n")
        config[CONF_API_ID] = int(input("Input an api id:\n>"))
        config[CONF_API_HASH] = str(input("Input an api hash:\n>"))

        safe_config(config)
        print("\nFinished the configuration.")


class ClusterStatusArgparser(ArgumentParser):
    """Class that deals with arguments for the telegram_simulation_bot"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_argument(
            "-c",
            "--config",
            action='store_true',
            help="start the config",
        )
        self.add_argument(
            "-d",
            "--delete",
            action='store_true',
            help="remove your config",
        )
        self.add_argument(
            "-t",
            "--type",
            default="MITgcm",
            help="specify which type of simulation you want to monitor",
        )



from .MITgcm import MITgcmHandler
from .config import ClusterStatusArgparser, read_config, CONF_API_ID, CONF_API_HASH, CONF_BOT_TOKEN, setup
from .handler import Handler

AVAILABLE_HANDLERS = ["mitgcm"]



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='telegram_simulation_bot',
    version='v0.1',
    packages=find_packages(),
    include_package_data=True,
    url='https://github.com/AaronDavidSchneider/telegram_simulation_bot',
    license='MIT',
    author='Aaron David Schneider',
    author_email='aarondavid.schneider@nbi.ku.dk',
    description='Interact with your computing cluster via telegram to get realtime information on your running jobs.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    scripts=['bin/telegram_simulation_bot'],
    install_requires=[
        "f90nml>=1.3.1",
        "telethon",
        "appdirs",
        "pyyaml",
    ]
)
